"""A tool for authoring Sphinx themes with a simple (opinionated) workflow.

This module also serves as the build-backend API for this project.
"""

__version__ = "0.2.0b2"

import contextlib
import sys
import tempfile
from pathlib import Path
from typing import Dict, Iterator, Optional

import rich

from ._internal.distributions import generate_metadata as _generate_metadata
from ._internal.distributions import generate_source_distribution as _generate_sdist
from ._internal.distributions import generate_wheel_distribution as _generate_wheel
from ._internal.errors import DiagnosticError
from ._internal.project import Project


@contextlib.contextmanager
def _ensure_has_metadata(
    project: Project, metadata_directory: Optional[str]
) -> Iterator[Path]:
    """Ensure that the body of a with statement, is executed with metadata generated."""
    if metadata_directory is not None:
        yield Path(metadata_directory)
    else:
        # The build frontend didn't generate the metadata. Generate it in a temporary
        # directory, and provide that. Cleans up when the context is exited.
        with tempfile.TemporaryDirectory() as dirname:
            metadata_dirname = Path(dirname)
            metadata_basename = _generate_metadata(
                project,
                destination=Path(dirname),
            )
            yield metadata_dirname / metadata_basename


@contextlib.contextmanager
def _clean_error_presentation() -> Iterator[None]:
    try:
        yield
    except DiagnosticError as error:
        rich.get_console().print(error, soft_wrap=True)
        sys.exit(1)


#
# Source Distributions
#
def build_sdist(
    sdist_directory: str,
    config_settings: Optional[Dict[str, str]] = None,
) -> str:
    """Generate a source distribution and place it into `sdist_directory`.

    https://www.python.org/dev/peps/pep-0517/#build-sdist
    """
    with _clean_error_presentation():
        project = Project.from_cwd()
        return _generate_sdist(project, destination=Path(sdist_directory))


#
# Wheel Distributions
#
def prepare_metadata_for_build_wheel(
    metadata_directory: str,
    config_settings: Optional[Dict[str, str]] = None,
) -> str:
    """Generate the metadata (.dist-info) and place it in `metadata_directory`.

    https://www.python.org/dev/peps/pep-0517/#prepare-metadata-for-build-wheel
    """
    with _clean_error_presentation():
        project = Project.from_cwd()
        return _generate_metadata(project, destination=Path(metadata_directory))


def build_wheel(
    wheel_directory: str,
    config_settings: Optional[Dict[str, str]] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Generate a wheel distribution and place it into `wheel_directory`.

    https://www.python.org/dev/peps/pep-0517/#build-wheel
    """
    with _clean_error_presentation():
        project = Project.from_cwd()
        with _ensure_has_metadata(project, metadata_directory) as metadata_path:
            return _generate_wheel(
                project,
                destination=Path(wheel_directory),
                metadata_directory=metadata_path,
                editable=False,
            )


#
# Editable installs
#
prepare_metadata_for_build_editable = prepare_metadata_for_build_wheel


def build_editable(
    wheel_directory: str,
    config_settings: Optional[Dict[str, str]] = None,
    metadata_directory: Optional[str] = None,
) -> str:
    """Generate an editable wheel distribution and place it into `wheel_directory`.

    https://www.python.org/dev/peps/pep-0660/#build-editable
    """
    with _clean_error_presentation():
        project = Project.from_cwd()
        with _ensure_has_metadata(project, metadata_directory) as metadata_path:
            return _generate_wheel(
                project,
                destination=Path(wheel_directory),
                metadata_directory=metadata_path,
                editable=True,
            )



"""Compilation entrypoint."""

import subprocess
import sys
from typing import List

import click

from ..ui import log


class PackageCommand:
    """Generate Python distribution files. (sdist and wheel)"""

    interface: List[click.Parameter] = []

    def run(self) -> int:
        """Make it happen."""
        log("[magenta]$[/] [blue]python -m build[/]")
        try:
            subprocess.run([sys.executable, "-m", "build"], check=True)
        except subprocess.CalledProcessError as error:
            return error.returncode
        log("[green]Done![/]")
        return 0



import inspect
import sys
from typing import Any, Dict, List, Optional, Protocol, TextIO, Type

import rich
from rich.text import Text

from ..errors import DiagnosticError

try:
    import build  # noqa
    import click
    import sphinx_autobuild  # type: ignore  # noqa
except ImportError as import_error:
    rich.print(
        DiagnosticError(
            reference="missing-command-line-dependencies",
            message=(
                "Could not import a package that is required for the `stb` command line."
            ),
            context=str(import_error),
            note_stmt=(
                "The optional [blue]cli[/] dependencies of this package are missing."
            ),
            hint_stmt=(
                "During installation, make sure to include the `\\[cli]`. For example:\n"
                'pip install "sphinx-theme-builder\\[cli]"'
            ),
        ),
        file=sys.stderr,
    )
    sys.exit(1)


class Command(Protocol):
    context_settings: Dict[str, Any]
    interface: List[click.Parameter]

    def run(self, **kwargs: Dict[str, Any]) -> int: ...


def create_click_command(cls: Type[Command]) -> click.Command:
    # Use the class docstring as the help string
    help_string = inspect.cleandoc(cls.__doc__)
    # Infer the name, from the known context.
    name = cls.__name__[: -len("Command")].lower()

    # Double check that things are named correctly.
    assert name.capitalize() + "Command" == cls.__name__
    assert name == cls.__module__.split(".")[-1]

    context_settings: Optional[Dict[str, Any]] = None
    if hasattr(cls, "context_settings"):
        context_settings = cls.context_settings

    command = click.Command(
        name=name,
        context_settings=context_settings,
        help=help_string,
        params=cls.interface,
        callback=lambda **kwargs: cls().run(**kwargs),
    )
    return command


def _root_callback() -> None:
    ctx = click.get_current_context()
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


def compose_command_line() -> click.Group:
    from .compile import CompileCommand
    from .new import NewCommand
    from .npm import NpmCommand
    from .package import PackageCommand
    from .serve import ServeCommand

    command_classes: List[Type[Command]] = [
        CompileCommand,  # type: ignore
        NewCommand,  # type: ignore
        PackageCommand,  # type: ignore
        ServeCommand,  # type: ignore
        NpmCommand,  # type: ignore
    ]

    # Convert our commands into click objects.
    click_commands = [create_click_command(cls) for cls in command_classes]

    # Create the main click interface.
    cli = click.Group(
        name="stb",
        help="sphinx-theme-builder helps make it easier to write sphinx themes.",
        commands={command.name: command for command in click_commands},  # type: ignore
        invoke_without_command=True,
        callback=_root_callback,
    )
    return cli


def present_click_usage_error(error: click.UsageError, *, stream: TextIO) -> None:
    assert error.ctx

    rich.print(
        Text.from_markup("[red]error:[/] ") + Text(error.format_message()),
        file=stream,
    )

    # Usage
    usage_parts = error.ctx.command.collect_usage_pieces(error.ctx)
    usage = " ".join([error.ctx.command_path] + usage_parts)
    print(file=stream)
    print("Usage:", file=stream)
    print(" ", usage, file=stream)

    # --help
    option = "--help"
    command = error.ctx.command_path
    print(file=stream)
    rich.print(
        f"Try [green]{command} {option}[/] for more information.",
        file=stream,
    )


def main(args: Optional[List[str]] = None) -> None:
    """The entrypoint for command line stuff."""
    cli = compose_command_line()
    try:
        cli(args=args, standalone_mode=False)
    except click.Abort:
        rich.print(r"[cyan]\[stb][/] [red]Aborting![/]", file=sys.stderr)
        sys.exit(1)
    except click.UsageError as error:
        present_click_usage_error(error, stream=sys.stderr)
        sys.exit(error.exit_code)  # uses exit codes 1 and 2
    except click.ClickException as error:
        error.show(sys.stderr)
        sys.exit(error.exit_code)  # uses exit codes 1 and 2
    except DiagnosticError as error:
        rich.print(error, file=sys.stderr)
        sys.exit(3)
    except Exception:
        console = rich.console.Console(stderr=True)
        console.print_exception(
            width=console.width, show_locals=True, word_wrap=True, suppress=[click]
        )
        console.print(
            DiagnosticError(
                reference="crash",
                message="A fatal error occurred.",
                context="See above for a detailed Python traceback.",
                note_stmt=(
                    "If you file an issue, please include the full traceback above."
                ),
                hint_stmt=(
                    "This might be due to an issue in sphinx-theme-builder, one of the "
                    "tools it uses internally, or your code."
                ),
            ),
        )
        sys.exit(4)



import os
import re
import subprocess
import sys
import tempfile
from pathlib import Path

import click

from ..errors import DiagnosticError
from ..project import Project
from ..ui import log


class ServeCommand:
    """Serve the provided documentation, with livereload on changes."""

    interface = [
        click.Option(
            ["--builder"],
            type=click.Choice(["html", "dirhtml"]),
            default="html",
            show_default=True,
            show_choices=True,
            help="The Sphinx builder to build the documentation with",
        ),
        click.Option(
            ["--host"],
            default=None,
            show_default=True,
            show_choices=True,
            help="hostname to serve documentation on (default: 127.0.0.1)",
        ),
        click.Option(
            ["--port"],
            type=click.INT,
            default=0,
            show_default=True,
            show_choices=True,
            help="The port to start the server on. Uses a random free port by default.",
        ),
        click.Option(
            ["--open-browser / --no-open-browser"],
            is_flag=True,
            default=True,
            show_default=True,
            help="Open the browser after starting live-reload server.",
        ),
        click.Option(
            ["--pdb"],
            is_flag=True,
            help="Run pdb if the Sphinx build fails with an exception.",
        ),
        click.Option(
            ["--override-theme / --no-override-theme"],
            is_flag=True,
            default=False,
            help="Override the `html_theme` value set in `conf.py`.",
        ),
        click.Argument(
            ["source_directory"],
            required=True,
            type=click.Path(
                exists=True,
                file_okay=False,
                dir_okay=True,
                allow_dash=True,
                resolve_path=True,
                path_type=Path,
            ),
        ),
    ]

    def run(
        self,
        *,
        host: str,
        port: int,
        source_directory: Path,
        builder: str,
        open_browser: bool,
        pdb: bool,
        override_theme: bool,
    ) -> int:
        project = Project.from_cwd()

        with tempfile.TemporaryDirectory() as build_directory:
            command = [
                sys.executable,
                "-m",
                "sphinx_autobuild",
                # sphinx-autobuild flags
                f"--watch={os.fsdecode(project.source_path)}",
                f"--re-ignore=({'|'.join(map(re.escape, project.compiled_assets))})",
                f"--port={port}",  # use a random free port
                f"--pre-build='{sys.executable}' -m sphinx_theme_builder compile",
                # Sphinx flags
                "-T",
                "-a",  # full rebuild to ensure static assets are copied on each change
                f"-b={builder}",
                os.fsdecode(source_directory),
                os.fsdecode(build_directory),
            ]
            if host:
                command.append(f"--host={host}")
            if pdb:
                command.append("-P")
            if open_browser:
                # open the browser for the user
                command.append("--open-browser")
            if override_theme:
                # override the theme, set in conf.py
                command.extend(["-D", f"html_theme={project.kebab_name}"])

            log("Invoking [bold]sphinx-autobuild[/] to run.")
            try:
                subprocess.run(command, check=True)
            except subprocess.CalledProcessError as error:
                raise DiagnosticError(
                    reference="autobuild-failed",
                    message=f"[b]sphinx-autobuild[/] exited with a non-zero exit code {error.returncode}.",
                    context="See above for failure output from the underlying tooling.",
                    hint_stmt=None,
                ) from error
        return 0



"""Used to initialise the project, with the correct scaffolding."""

import os
import subprocess
import sys
from pathlib import Path
from typing import List

import click

from ..errors import DiagnosticError
from ..ui import log

_TEMPLATE_URL = "https://github.com/pradyunsg/sphinx-theme-template/archive/refs/heads/sphinx-theme-template-stable.zip"


class NewCommand:
    """Create a new project."""

    interface: List[click.Parameter] = [
        click.Argument(
            ["source_directory"],
            required=True,
            type=click.Path(
                exists=False,
                file_okay=False,
                dir_okay=True,
                writable=True,
                path_type=Path,
            ),
        ),
    ]

    def run(self, source_directory: Path) -> int:
        log(f"[yellow]#[/] Looking at [magenta]{source_directory}[/]")

        pyproject_toml = source_directory / "pyproject.toml"
        setup_py = source_directory / "setup.py"
        if pyproject_toml.exists() or setup_py.exists():
            raise DiagnosticError(
                reference="can-not-overwrite-existing-python-project",
                message="Refusing to generate a new project in provided directory.",
                context=(
                    "This directory contains a Python project, which will not be "
                    "overwritten."
                ),
                hint_stmt=(
                    "This command should be used on empty/non-existing directories."
                ),
            )

        command = [
            sys.executable,
            "-m",
            "cookiecutter",
            "-o",
            os.fsdecode(source_directory),
            _TEMPLATE_URL,
        ]
        log(f"[magenta]$[/] python {' '.join(command[1:])}")
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as error:
            raise DiagnosticError(
                reference="cookiecutter-failed",
                message="Cookiecutter failed to generate the project.",
                context=f"Exit code: {error.returncode}",
                note_stmt="cookiecutter's failure output is available above.",
                hint_stmt=(
                    "[b]@pradyunsg[/] might still need to write/update the supporting "
                    "cookiecutter template. :sweat_smile:"
                ),
            ) from error

        return 0



"""Logic for generating distributions from a :ref:`Project`."""

import io
import os
import posixpath
import subprocess
import tarfile
import textwrap
from pathlib import Path
from typing import Callable, Optional, Tuple

from .errors import DiagnosticError
from .nodejs import generate_assets
from .project import Project
from .wheelfile import WheelFile, include_parent_paths


def _get_vcs_tracked_names(path: Path) -> Optional[Tuple[str, ...]]:
    if not (path / ".git").is_dir():
        return None

    outb = subprocess.check_output(
        ["git", "ls-files", "--recurse-submodules", "-z"],
        cwd=path,
    )

    tracked_files = [
        os.fsdecode(location) for location in outb.strip(b"\0").split(b"\0") if location
    ]
    return include_parent_paths(tracked_files)


def _sdist_filter(
    project: Project, base: str
) -> Callable[[tarfile.TarInfo], Optional[tarfile.TarInfo]]:
    """Create a filter to pass to tarfile.add, for this project."""
    compiled_assets = project.compiled_assets
    tracked_names = _get_vcs_tracked_names(project.location)

    def _filter(tarinfo: tarfile.TarInfo) -> Optional[tarfile.TarInfo]:
        # Include the entry for the root.
        if tarinfo.name == base:
            return tarinfo

        name = tarinfo.name[len(base) + 1 :]

        # Exclude hidden files
        if name.startswith("."):
            return None

        # Exclude compiled pyc files.
        if posixpath.basename(name) == "__pycache__":
            return None

        # Exclude compiled assets.
        if name in compiled_assets:
            return None

        # Exclude things that are excluded from version control.
        if tracked_names is not None and name not in tracked_names:
            return None

        return tarinfo

    # NOTE: The CLI for build needs to check that the compiled assets we have here, are
    #       not tracked under version control.
    return _filter


#
# External API
#
def generate_source_distribution(
    project: Project,
    *,
    destination: Path,
) -> str:
    """Generate a source distribution for project, and place it inside destination.

    :return: Name of the generated source tarball.
    """
    os.makedirs(destination, exist_ok=True)

    dashed_pair = f"{project.snake_name}-{project.metadata.version}"
    sdist_tarball = destination / f"{dashed_pair}.tar.gz"

    # NOTE: Post Python 3.7 -- can drop the format kwarg, since it'll be the default.
    with tarfile.open(
        name=sdist_tarball, mode="w:gz", format=tarfile.PAX_FORMAT
    ) as tarball:
        # Recursively add the files to this tarball, with an exclusion filter.
        tarball.add(
            project.location,
            arcname=dashed_pair,
            recursive=True,
            filter=_sdist_filter(project, dashed_pair),
        )

        # Write the metadata file.
        metadata_content = project.get_metadata_file_contents().encode()
        tarinfo = tarfile.TarInfo(posixpath.join(dashed_pair, "PKG-INFO"))
        tarinfo.size = len(metadata_content)
        tarball.addfile(tarinfo, io.BytesIO(metadata_content))

    return sdist_tarball.name


def generate_metadata(
    project: Project,
    *,
    destination: Path,
) -> str:
    """Generate the metadata (.dist-info) and place it in `metadata_directory`.

    :return: name of the dist-info directory generated.
    """
    dist_info = (
        destination / f"{project.snake_name}-{project.metadata.version}.dist-info"
    )
    try:
        os.makedirs(dist_info)
    except OSError as error:
        raise DiagnosticError(
            message="Metadata directory already exists",
            context=None,
            hint_stmt=None,
        ) from error

    # Delegated generation.
    (dist_info / "entry_points.txt").write_text(project.get_entry_points_contents())
    (dist_info / "LICENSE").write_text(project.get_license_contents())
    (dist_info / "METADATA").write_text(project.get_metadata_file_contents())

    # Templated generation.
    (dist_info / "WHEEL").write_text(
        textwrap.dedent(
            """\
            Wheel-Version: 1.0
            Generator: sphinx-theme-builder {version}
            Root-Is-Purelib: true
            Tag: py3-none-any
            """
        )
    )

    return dist_info.name


def generate_wheel_distribution(
    project: Project,
    *,
    destination: Path,
    metadata_directory: Path,
    editable: bool,
) -> str:
    # Generate the JS / CSS assets
    generate_assets(project, production=not editable)

    wheel_path = (
        destination
        / f"{project.snake_name}-{project.metadata.version}-py3-none-any.whl"
    )
    tracked_names = _get_vcs_tracked_names(project.location)

    with WheelFile(
        path=wheel_path,
        compiled_assets=project.compiled_assets,
        tracked_names=tracked_names,
    ) as wheel:
        if editable:
            # Generate a .pth file, for editable installs. There's an enforced src/
            # directory, so this is fairly safe to do.
            wheel.add_string(
                os.fsdecode(project.source_path.resolve()),
                dest=project.snake_name + ".pth",
            )
        else:
            # Add files from the project's src/ directory.
            wheel.add_directory(project.source_path, dest="", base=project.location)

        # Put the metadata at the end.
        # https://www.python.org/dev/peps/pep-0427/#recommended-archiver-features
        wheel.add_directory(metadata_directory, dest=metadata_directory.name, base=None)
        wheel.write_record(dest=metadata_directory.name + "/RECORD")

    return wheel.name



"""Compilation entrypoint."""

from typing import List

import click

from ..nodejs import generate_assets
from ..project import Project


class CompileCommand:
    """Compile the current project's assets."""

    interface: List[click.Parameter] = [
        click.Option(
            ["--production"],
            is_flag=True,
            default=False,
            help="Runs the build with `NODE_ENV=production` (`development` by default).",
        ),
    ]

    def run(self, production: bool) -> int:
        """Make it happen."""
        project = Project.from_cwd()
        generate_assets(project, production=production)
        return 0



"""A helper class to make it easier to generate a wheel.
"""

import base64
import hashlib
import os
import posixpath
import zipfile
from dataclasses import dataclass
from pathlib import Path, PurePosixPath
from types import TracebackType
from typing import BinaryIO, List, Optional, Set, Tuple, Type

_HASH_ALGORITHM = "sha256"

# Borrowed from CPython's shutil.py
# https://github.com/python/cpython/blob/v3.9.1/Lib/shutil.py#L52
_WINDOWS = os.name == "nt"
_COPY_BUFSIZE = 1024 * 1024 if _WINDOWS else 64 * 1024


def encode_for_record(hasher: "hashlib._Hash") -> str:
    """Takes a hasher and returns the encoded hash value for the RECORD."""
    return base64.urlsafe_b64encode(hasher.digest()).decode("ascii").rstrip("=")


# Adapted from pradyunsg/installer
# https://github.com/pradyunsg/installer/blob/0.7.0/src/installer/utils.py#L95
def copyfileobj_with_hashing(
    source: BinaryIO,
    dest: BinaryIO,
    hash_algorithm: str,
) -> Tuple[str, int]:
    """Copy a buffer while computing the content's hash and size.

    Copies the source buffer into the destination buffer while computing the
    hash of the contents. Adapted from :any:`shutil.copyfileobj`.

    :param source: buffer holding the source data
    :param dest: destination buffer
    :param hash_algorithm: hashing algorithm

    :return: size, hash digest of the contents
    """
    hasher = hashlib.new(hash_algorithm)
    size = 0
    while True:
        buf = source.read(_COPY_BUFSIZE)
        if not buf:
            break
        hasher.update(buf)
        dest.write(buf)
        size += len(buf)

    return encode_for_record(hasher), size


def include_parent_paths(posix_style_paths: List[str]) -> Tuple[str, ...]:
    names: Set[str] = set()
    for path_str in posix_style_paths:
        path = PurePosixPath(path_str)
        names.update(parent.as_posix() for parent in path.parents)
        names.add(path_str)
    return tuple(names)


@dataclass
class RecordEntry:
    """A single entry in a RECORD file."""

    path: str
    hash_value: str
    hash_algorithm: str
    size: str

    def to_line(self) -> str:
        if self.hash_value:
            return ",".join(
                (self.path, f"{self.hash_algorithm}={self.hash_value}", self.size)
            )
        else:
            assert not self.hash_algorithm
            return ",".join((self.path, "", self.size))


class WheelFile:
    """A helper class to generate wheels."""

    def __init__(
        self,
        *,
        path: Path,
        tracked_names: Optional[Tuple[str, ...]],
        compiled_assets: Tuple[str, ...],
    ) -> None:
        self._path = path
        self._tracked_names = tracked_names
        self._compiled_assets = compiled_assets
        self._compiled_assets_and_parents = include_parent_paths(list(compiled_assets))
        self._zipfile = zipfile.ZipFile(path, mode="w")
        self._records: List[RecordEntry] = []

    def __enter__(self) -> "WheelFile":
        return self

    def __exit__(
        self,
        type_: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._zipfile.close()

    def _exclude(self, path: Path, *, base: Optional[Path]) -> bool:
        # Exclude compiled pyc files.
        if path.name == "__pycache__":
            return True

        # Include all not-based-on-source-tree files.
        if base is None:
            return False

        normalised_path = path.relative_to(base).as_posix()

        # Definitely include compiled assets.
        if normalised_path in self._compiled_assets_and_parents:
            return False

        for asset_path in self._compiled_assets:
            if normalised_path.startswith(asset_path + "/"):
                return False

        # Exclude things that are excluded from version control.
        if self._tracked_names is not None:
            if normalised_path not in self._tracked_names:
                return True

        # If we're here, we've excluded all the files that needed to be excluded.
        return False

    @property
    def name(self) -> str:
        """Name of the wheel."""
        return self._path.name

    def write_record(self, *, dest: str) -> None:
        """Write the record, in the provided destination.

        :param dest: The exact ``{package}-{version}.dist-info/RECORD`` to write to.
        """
        assert self._zipfile.fp is not None

        self._records.append(RecordEntry(dest, "", "", ""))
        lines = [record.to_line() for record in self._records]
        self._zipfile.writestr(dest, data="\n".join(lines))

    def add_string(self, content: str, *, dest: str) -> None:
        """Add a file at ``dest``, with the given ``content``."""
        assert self._zipfile.fp is not None

        data = content.encode()
        hasher = hashlib.new(_HASH_ALGORITHM, data=data)

        self._zipfile.writestr(dest, data=data)
        self._records.append(
            RecordEntry(
                path=dest,
                hash_algorithm=_HASH_ALGORITHM,
                hash_value=encode_for_record(hasher),
                size=str(len(data)),
            )
        )

    def add_file(self, file: Path, *, dest: str, base: Optional[Path]) -> None:
        """Add a file at ``dest``, with the contents of ``file``."""
        assert self._zipfile.fp is not None

        if self._exclude(file, base=base):
            return

        # Copy the file object.
        zipinfo = zipfile.ZipInfo.from_file(file, dest)
        with file.open("rb") as source_stream:
            with self._zipfile.open(zipinfo, "w") as dest_stream:
                hash_value, size = copyfileobj_with_hashing(
                    source_stream, dest_stream, hash_algorithm=_HASH_ALGORITHM
                )
        self._records.append(
            RecordEntry(
                path=dest,
                hash_algorithm=_HASH_ALGORITHM,
                hash_value=hash_value,
                size=str(size),
            )
        )

    def add_directory(
        self, directory: Path, *, dest: str, base: Optional[Path]
    ) -> None:
        """Add the directory to the archive, recursively."""
        assert directory.is_dir()
        assert self._zipfile.fp is not None

        if self._exclude(directory, base=base):
            return

        for item in sorted(directory.iterdir()):
            if item.is_dir():
                self.add_directory(
                    item, dest=posixpath.join(dest, item.name), base=base
                )
            else:
                self.add_file(item, dest=posixpath.join(dest, item.name), base=base)



"""NodeJS tooling orchestration.

Broadly, it has a `.nodeenv` created using the nodeenv package; and ensures that the
`node_modules` directory is kept in sync with the `package.json` file of the project.
"""

import os
import shlex
import shutil
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import Any, List, Optional
from unittest.mock import patch

from rich.markup import escape

from .errors import DiagnosticError
from .passthrough import passthrough_run
from .project import Project
from .ui import log

_NODEENV_DIR = ".nodeenv"
_BIN_DIR = "Scripts" if os.name == "nt" else "bin"


def _get_bool_env_var(name: str, *, default: bool) -> bool:
    value = os.environ.get(name)
    if value is None:
        return default
    if value.lower() in {"false", "0"}:
        return False
    if value.lower() in {"true", "1"}:
        return True
    raise DiagnosticError(
        reference="non-boolean-env-variable-value",
        message=f"The provided value for `{name}` is invalid.",
        context=f"{name}={escape(value)}",
        hint_stmt=f"Provide a boolean value for `{name}` (true, false, 1, 0).",
    )


def run_in(
    nodeenv: Path, args: List[str], *, production: bool = False, **kwargs: Any
) -> "Optional[subprocess.CompletedProcess[bytes]]":
    """Run a command, using a binary from `nodeenv`."""
    assert nodeenv.name == _NODEENV_DIR

    log(f"[magenta](nodeenv) $[/] [blue]{' '.join(args)}[/]")
    env = {
        "NPM_CONFIG_PREFIX": os.fsdecode(nodeenv),
        "npm_config_prefix": os.fsdecode(nodeenv),
        "NODE_PATH": os.fsdecode(nodeenv / "lib" / "node_modules"),
        "PATH": os.pathsep.join([os.fsdecode(nodeenv / _BIN_DIR), os.environ["PATH"]]),
        "NODE_ENV": "production" if production else "development",
    }

    # Fully qualify the first argument.
    resolved_name = shutil.which(args[0], path=env["PATH"])
    if not resolved_name:
        raise FileNotFoundError(resolved_name)
    args[0] = resolved_name

    with patch.dict("os.environ", env):
        if not kwargs:
            returncode = passthrough_run(args)
            if returncode:
                cmd = " ".join(shlex.quote(arg) for arg in args)
                raise subprocess.CalledProcessError(returncode=returncode, cmd=cmd)
            return None
        else:
            return subprocess.run(args, check=True, **kwargs)


def _run_python_nodeenv(*args: str) -> None:
    presentation = ["python", "-m", "nodeenv", *args]
    log(f"[magenta]$[/] [blue]{' '.join(presentation)}[/]")

    command = [
        sys.executable,
        "-c",
        textwrap.dedent(
            """
            import runpy
            import rich
            import rich.traceback
            import urllib.request

            rich.traceback.install(
                width=rich.get_console().width,
                show_locals=True,
                suppress=[runpy, urllib.request],
            )
            runpy.run_module("nodeenv", run_name="__main__", alter_sys=True)
            """
        ),
        "-v",
        *args,
    ]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as error:
        raise DiagnosticError(
            reference="nodeenv-creation-failed",
            message="Failed to create a `nodeenv`",
            context="See above for failure output from the underlying tooling.",
            hint_stmt=(
                "A `urllib.error.HTTPError` indicates that the issue is "
                "related to the network or the availability of NodeJS release files. "
                "It may mean the node version that this tool is trying to fetch is no "
                "longer available, for example if there is no compatible NodeJS binary "
                "for the operating system."
            ),
        ) from error


def _should_use_system_node(node_version: str) -> bool:
    if not _get_bool_env_var("STB_USE_SYSTEM_NODE", default=False):
        return False

    if sys.platform == "win32":
        raise DiagnosticError(
            reference="can-not-use-system-node-on-windows",
            message="sphinx-theme-builder can not use the system node on Windows.",
            context="The underlying tooling (nodeenv) does not yet support this.",
            hint_stmt="Unset `STB_USE_SYSTEM_NODE`, which is currently set to 'true'",
        )
    return True


def _relaxed_version_check(expected: str, got: str) -> Optional[str]:
    """Perform a relaxed check of the failure mode.

    Returns None, if the versions match the relaxed check criterion.
    Returns a string, with the reason for the the version check failure.
    """
    assert expected.startswith("v"), expected
    assert got.startswith("v"), got

    expected_parts = expected[1:].split(".")[:2]
    got_parts = got[1:].split(".")[:2]

    assert len(expected_parts) == len(got_parts) == 2, f"{expected};{got}"

    if expected_parts[0] != got_parts[0]:
        return "Minor version does not match the value declared in the theme."
    if int(expected_parts[1]) < int(got_parts[1]):
        return "Minor version is lower than the value declared in the theme."
    return None


def ensure_version_matches(expected: str, got: str) -> None:
    """Ensure that the version of NodeJS matches the expected value.

    STB_RELAX_NODE_VERSION_CHECK can be used to relax the check, to only check the major
    and minor version.
    """
    if got == expected:
        return

    if not _get_bool_env_var("STB_RELAX_NODE_VERSION_CHECK", default=False):
        raise DiagnosticError(
            reference="nodeenv-version-mismatch",
            message="The `nodeenv` for this project is unhealthy.",
            context=(
                "There is a mismatch between what is present in the environment "
                f"({got}) and the expected version of NodeJS ({expected})."
            ),
            hint_stmt=(
                f"Deleting the {_NODEENV_DIR} directory and trying again may work."
            ),
        )

    log("[yellow]#[/] The node version is not the expected one.")
    log("[yellow]#[/] The generated assets may be broken even if the build succeeds.")
    log("[yellow]#[/] Continuing anyway - `STB_RELAX_NODE_VERSION_CHECK` is truthy.")

    rejection_reason = _relaxed_version_check(expected, got)
    if rejection_reason is None:
        return

    raise DiagnosticError(
        reference="node-version-mismatch",
        message="The node version is not compatible with the expected version.",
        context=f"{rejection_reason}\nSee above for the expected and actual version.",
        hint_stmt=(
            "You need to use a compatible version of NodeJS to build the theme. "
        ),
    )


def create_nodeenv(nodeenv: Path, node_version: str) -> None:
    """Create a `nodeenv` for the theme."""
    log(
        "[yellow]#[/] [cyan]Generating new [magenta]nodeenv[/] with "
        f"NodeJS [green]{node_version}[/]!"
    )

    if _should_use_system_node(node_version=node_version):
        log("[yellow]#[/] Will use system nodeJS.")
        node_version = "system"

    envdir = os.fsdecode(nodeenv)
    # This next bit is borrowed from
    # https://github.com/pre-commit/pre-commit/blob/v2.16.0/pre_commit/languages/node.py
    if sys.platform == "win32":  # pragma: no cover
        envdir = "\\\\?\\" + os.path.normpath(envdir)

    _run_python_nodeenv(
        f"--node={node_version}",
        "--prebuilt",
        "--clean-src",
        envdir,
    )


def run_npm_build(nodeenv: Path, *, production: bool) -> None:
    try:
        run_in(nodeenv, ["npm", "run-script", "build"], production=production)
    except subprocess.CalledProcessError as error:
        raise DiagnosticError(
            reference="js-build-failed",
            message="The Javascript-based build pipeline failed.",
            context="See above for failure output from the underlying tooling.",
            hint_stmt=None,
        ) from error


def populate_npm_packages(nodeenv: Path, node_modules: Path) -> None:
    try:
        run_in(nodeenv, ["npm", "install", "--include=dev"])
    except FileNotFoundError as error:
        raise DiagnosticError(
            reference="nodeenv-unhealthy-npm-not-found",
            message="The `nodeenv` for this project is unhealthy.",
            context=str(error),
            hint_stmt=(
                f"Deleting the {_NODEENV_DIR} directory and trying again may work."
            ),
        ) from error
    except subprocess.CalledProcessError as error:
        raise DiagnosticError(
            reference="js-install-failed",
            message="Javascript dependency installation failed.",
            context="See above for failure output from the underlying tooling.",
            hint_stmt=None,
        ) from error

    if node_modules.is_dir():
        node_modules.touch()


def generate_assets(project: Project, *, production: bool) -> None:
    package_json = project.location / "package.json"
    nodeenv = project.location / _NODEENV_DIR
    node_modules = project.location / "node_modules"

    assert package_json.exists()

    created_new_nodeenv = False
    if not nodeenv.exists():
        log("[yellow]#[/] [magenta]nodeenv[cyan] does not exist.[/]")
        create_nodeenv(nodeenv, project.node_version)
        created_new_nodeenv = True

    # Checking the node version is a sanity check, and ensures that the environment is
    # "healthy".
    try:
        process = run_in(nodeenv, ["node", "--version"], stdout=subprocess.PIPE)
    except FileNotFoundError as error:
        raise DiagnosticError(
            reference="nodeenv-unhealthy-file-not-found",
            message="The `nodeenv` for this project is unhealthy.",
            context=str(error),
            hint_stmt=(
                f"Deleting the {_NODEENV_DIR} directory and trying again may work."
            ),
        ) from error
    except subprocess.CalledProcessError as error:
        raise DiagnosticError(
            reference="nodeenv-unhealthy-subprocess-failure",
            message="The `nodeenv` for this project is unhealthy.",
            context="See above for failure output from the underlying tooling.",
            hint_stmt=(
                f"Deleting the {_NODEENV_DIR} directory and trying again may work."
            ),
        ) from error

    # Present the `node --version` output to the user.
    assert process
    got = process.stdout.decode().strip()
    print(got)

    # Sanity-check the node version.
    expected = f"v{project.node_version}"
    ensure_version_matches(expected, got)

    need_to_populate = False
    if created_new_nodeenv:
        need_to_populate = True
    elif not node_modules.exists():
        need_to_populate = True
    elif node_modules.stat().st_mtime < package_json.stat().st_mtime:
        log("[yellow]#[/] [cyan]Detected changes in [magenta]package.json[cyan].[/]")
        need_to_populate = True

    if need_to_populate:
        if node_modules.exists():
            log("[yellow]#[/] [cyan]Cleaning up [magenta]node_modules[cyan].[/]")
            try:
                shutil.rmtree(node_modules)
            except OSError as error:
                raise DiagnosticError(
                    reference="unable-to-cleanup-node-modules",
                    message="Could not remove node_modules directory.",
                    context=str(error),
                    hint_stmt=f"Deleting {node_modules} and trying again may work.",
                ) from error

        log("[yellow]#[/] [cyan]Installing NodeJS packages.[/]")
        populate_npm_packages(nodeenv, node_modules)

    run_npm_build(nodeenv=nodeenv, production=production)

    log("[green]Done![/]")



"""Used to initialise the project, with the correct scaffolding."""

import subprocess
from pathlib import Path
from typing import List, Tuple

import click

from ..errors import DiagnosticError
from ..nodejs import _NODEENV_DIR as NODEENV_DIRNAME
from ..nodejs import run_in


class NpmCommand:
    """Interact with the npm, available in the environment.

    If you want to run `npm --help`, use `--` to separate the options for npm:

        stb npm -- --help
    """

    context_settings = dict(
        ignore_unknown_options=True,
    )
    interface: List[click.Parameter] = [
        click.Argument(["arguments"], nargs=-1, type=click.UNPROCESSED),
    ]

    def run(self, arguments: Tuple[str]) -> int:
        nodeenv = Path.cwd() / NODEENV_DIRNAME
        if not nodeenv.exists():
            raise DiagnosticError(
                message="Could not find the `.nodeenv` directory to use.",
                context=str(nodeenv),
                hint_stmt="Did you run `stb compile` yet?",
                reference="no-nodeenv",
            )

        try:
            run_in(nodeenv, ["npm"] + list(arguments))
        except subprocess.CalledProcessError as error:
            return error.returncode
        else:
            return 0



"""Exceptions raised from within this package."""

import re
from typing import TYPE_CHECKING, Optional, Union

from rich.console import Console, ConsoleOptions, RenderResult
from rich.text import Text

if TYPE_CHECKING:
    from typing import Literal

_DOCS_URL = "https://sphinx-theme-builder.rtfd.io/errors/#{}"


def _is_kebab_case(s: str) -> bool:
    return re.match(r"^[a-z]+(-[a-z]+)*$", s) is not None


def _prefix_with_indent(
    s: Union[Text, str],
    console: Console,
    *,
    width_offset: int = 0,
    prefix: str,
    indent: str,
) -> Text:
    if isinstance(s, Text):
        text = s
    else:
        text = console.render_str(s)

    lines = text.wrap(console, console.width - width_offset)

    return console.render_str(prefix) + console.render_str(f"\n{indent}").join(lines)


class DiagnosticError(Exception):
    reference: str

    def __init__(
        self,
        *,
        kind: 'Literal["error", "warning"]' = "error",
        reference: Optional[str] = None,
        message: Union[str, Text],
        context: Optional[Union[str, Text]],
        hint_stmt: Optional[Union[str, Text]],
        note_stmt: Optional[Union[str, Text]] = None,
    ) -> None:
        # Ensure a proper reference is provided.
        if reference is None:
            assert hasattr(self, "reference"), "error reference not provided!"
            reference = self.reference
        assert _is_kebab_case(reference), "error reference must be kebab-case!"

        self.kind = kind
        self.reference = reference

        self.message = message
        self.context = context

        self.note_stmt = note_stmt
        self.hint_stmt = hint_stmt

        super().__init__(
            f"<{self.__class__.__name__}: {_DOCS_URL.format(self.reference)}>"
        )

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}("
            f"reference={self.reference!r}, "
            f"message={self.message!r}, "
            f"context={self.context!r}, "
            f"note_stmt={self.note_stmt!r}, "
            f"hint_stmt={self.hint_stmt!r}"
            ")>"
        )

    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        colour = "red" if self.kind == "error" else "yellow"

        yield f"[{colour} bold]{self.kind}[/]: [bold]{self.reference}[/]"

        # Present the main message, with relevant context indented.
        if not options.ascii_only:
            yield ""
            if self.context is not None:
                yield _prefix_with_indent(
                    self.message,
                    console,
                    width_offset=2,
                    prefix=f"[{colour}]×[/] ",
                    indent=f"[{colour}]│[/] ",
                )
                yield _prefix_with_indent(
                    self.context,
                    console,
                    width_offset=4,
                    prefix=f"[{colour}]╰─>[/] ",
                    indent=f"[{colour}]   [/] ",
                )
            else:
                yield _prefix_with_indent(
                    self.message,
                    console,
                    width_offset=4,
                    prefix="[red]×[/] ",
                    indent="  ",
                )
        else:  # coverage: skip
            yield console.render_str(f"[{colour}]x[/] ") + self.message
            if self.context is not None:
                yield ""
                yield self.context

        yield ""

        if self.note_stmt is not None:
            yield _prefix_with_indent(
                self.note_stmt,
                console,
                width_offset=6,
                prefix="[magenta bold]note[/]: ",
                indent="      ",
            )
        if self.hint_stmt is not None:
            yield _prefix_with_indent(
                self.hint_stmt,
                console,
                width_offset=6,
                prefix="[cyan bold]hint[/]: ",
                indent="      ",
            )

        yield f"[bold]link[/]: {_DOCS_URL.format(self.reference)}"



"""Metadata and file system logic goes here."""

import ast
import configparser
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import pyproject_metadata
from packaging.utils import canonicalize_name
from packaging.version import InvalidVersion, Version
from rich.text import Text

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from .errors import DiagnosticError

if TYPE_CHECKING:
    from typing import Literal


def read_toml_file(path: Path) -> Dict[str, Any]:
    with path.open("rb") as stream:
        return tomllib.load(stream)


def get_version_using_ast(contents: bytes) -> Optional[str]:
    """Extract the version from the given file, using the Python AST."""
    tree = ast.parse(contents)

    # Only need to check the top-level nodes, and not recurse deeper.
    for child in tree.body:
        # Look for a simple string assignment to __version__
        if (
            isinstance(child, ast.Assign)
            and len(child.targets) == 1
            and isinstance(child.targets[0], ast.Name)
            and child.targets[0].id == "__version__"
            and isinstance(child.value, ast.Constant)
            and isinstance(child.value.value, str)
        ):
            return child.value.value

    return None


class ImproperProjectMetadata(DiagnosticError):
    """For issues with pyproject.toml contents or other metadata."""


class InvalidProjectStructure(DiagnosticError):
    """For issues with the project structure."""


def _load_pyproject(pyproject: Path) -> Tuple[str, Dict[str, Any]]:
    """Load from the pyproject.toml file, doing the minimal sanity checks."""
    try:
        pyproject_data = read_toml_file(pyproject)
    except FileNotFoundError as error:
        raise InvalidProjectStructure(
            message="Could not find a `pyproject.toml`.",
            context=f"Looked at {pyproject}",
            hint_stmt="Is this a valid Python package?",
            reference="pyproject-missing",
        ) from error
    except tomllib.TOMLDecodeError as error:
        raise ImproperProjectMetadata(
            message="Could not parse `pyproject.toml`.",
            context=f"{error}\npath: {pyproject}",
            hint_stmt=None,
            reference="pyproject-could-not-parse",
        ) from error

    project = pyproject_data.get("project", None)
    if project is None:
        raise ImproperProjectMetadata(
            message=Text("Could not find [project] table."),
            context=f"in file {pyproject}",
            hint_stmt=None,
            reference="pyproject-no-project-table",
        )

    kebab_name = project.get("name", None)
    if kebab_name is None:
        raise ImproperProjectMetadata(
            message=Text("Could not find `name` in [project] table."),
            context=f"in file {pyproject}",
            hint_stmt=None,
            reference="pyproject-no-name-in-project-table",
        )

    canonical_name = canonicalize_name(kebab_name)
    if kebab_name != canonical_name:
        raise ImproperProjectMetadata(
            message=Text("Found non-canonical `name` declared in the [project] table."),
            context=(
                f"Got '{kebab_name}', expected '{canonical_name}'\nin file {pyproject}"
            ),
            hint_stmt=None,
            reference="pyproject-non-canonical-name",
        )

    return kebab_name, pyproject_data


def _determine_version(
    package_path: Path, kebab_name: str, pyproject_data: Dict[str, Any]
) -> 'Tuple[str, Literal["pyproject.toml", "__init__.py"]]':
    # Let's look for the version now!
    declared_in_python = None  # type: Optional[str]
    declared_in_pyproject = None  # type: Optional[str]

    metadata = pyproject_data["project"]

    # Load the version from pyproject.toml file, if provided.
    if "version" in metadata:
        declared_in_pyproject = metadata["version"]
        if not isinstance(declared_in_pyproject, str):
            raise ImproperProjectMetadata(
                message=Text(
                    "Expected 'version' in the [project] table to be a string."
                ),
                context=(
                    f"Got {declared_in_pyproject} which is {type(declared_in_pyproject)} instead."
                ),
                hint_stmt=None,
                reference="pyproject-invalid-version",
            )

    # Load the version from __init__ file, if provided.
    package_init_file = package_path / "__init__.py"
    if not package_init_file.exists():
        raise InvalidProjectStructure(
            message=f"{package_init_file} is missing.",
            context=(
                "It is required for using sphinx-theme-builder.\n"
                f"The declared project name is {kebab_name}."
            ),
            hint_stmt=None,
            reference="project-init-missing",
        )

    try:
        declared_in_python = get_version_using_ast(package_init_file.read_bytes())
    except SyntaxError as error:
        raise InvalidProjectStructure(
            message="Could not parse `__init__.py` file",
            context=f"In file {package_init_file}\n{error!r}",
            hint_stmt=None,
            reference="project-init-invalid-syntax",
        ) from error

    if declared_in_pyproject and declared_in_python:
        raise InvalidProjectStructure(
            message="Found version declaration in both `pyproject.toml` and `__init__.py`.",
            context=(
                "The package version MUST only be provided information in one location"
                f"\nFrom `pyproject.toml`, got {declared_in_pyproject}"
                f"\nFrom `__init__.py`, got {declared_in_python}"
            ),
            hint_stmt=Text(
                "It is a good idea to declare the version in Python alone.\n"
                "You can do this by removing `project.version` from `pyproject.toml` "
                'and including "version" in the `project.dynamic` list. Like so:\n\n'
                "[project]\n"
                'dynamic = ["version"]'
            ),
            reference="project-double-version-declaration",
        )
    elif declared_in_python:
        if "version" not in metadata.get("dynamic", []):
            raise ImproperProjectMetadata(
                message=(
                    "Found version in `__init__.py` but pyproject.toml does not "
                    'include "version" in `project.dynamic` list.'
                ),
                context=f"From `__init__.py`, got version {declared_in_python}",
                hint_stmt=Text(
                    'You solve this by including "version" in the `project.dynamic` '
                    "list. Like so:\n\n"
                    "[project]\n"
                    'dynamic = ["version"]'
                ),
                reference="project-missing-dynamic-version",
            )
        return (declared_in_python, "__init__.py")
    elif declared_in_pyproject:
        if "version" in metadata.get("dynamic", []):
            raise ImproperProjectMetadata(
                message=(
                    'Declared "version" as `dynamic`, while also using `version` key '
                    "in `pyproject.toml`."
                ),
                context=f"From `pyproject.toml`, got version {declared_in_pyproject}.",
                hint_stmt=(
                    "You can solve this by removing `version` from `project.dynamic`."
                ),
                reference="project-improper-dynamic-version",
            )
        return (declared_in_pyproject, "pyproject.toml")

    raise InvalidProjectStructure(
        message="No version declaration found for project.",
        context=f"Looked at {package_init_file} and corresponding `pyproject.toml`.",
        hint_stmt=(
            "Did you forget to declare the version? "
            "It's the 'project.version' key in `pyproject.toml` "
            "or the `__version__` attribute in `__init__.py`."
        ),
        reference="project-no-version-declaration",
    )


@dataclass(frozen=True)
class Project:
    """Represents a project to be built."""

    kebab_name: str
    location: Path
    metadata: pyproject_metadata.StandardMetadata

    theme_name: str
    node_version: str
    additional_compiled_static_assets: List[str]

    @classmethod
    def from_cwd(cls) -> "Project":
        retval = cls.from_path(Path.cwd())
        retval.validate_file_structure_and_contents()
        return retval

    @classmethod
    def from_path(cls, path: Path) -> "Project":
        """Load a project from given Path."""
        pyproject = path / "pyproject.toml"
        kebab_name, pyproject_data = _load_pyproject(pyproject)

        # IMPORTANT: Keep in sync with `python_package_path` below.
        package_path = path / "src" / kebab_name.replace("-", "_")

        # Get the version, and validate it.
        version_s, version_comes_from = _determine_version(
            package_path, kebab_name, pyproject_data
        )

        try:
            version = Version(version_s)
        except InvalidVersion as error:
            raise ImproperProjectMetadata(
                message=f"Got an invalid version: {version_s}",
                context=f"This came from `{version_comes_from}`",
                hint_stmt=None,
                reference="project-invalid-version",
            ) from error

        # Get the metadata, and validate it.
        try:
            metadata = pyproject_metadata.StandardMetadata.from_pyproject(
                pyproject_data, path
            )
        except pyproject_metadata.ConfigurationError as error:
            raise InvalidProjectStructure(
                message="Provided project metadata is not valid.",
                context=str(error),
                hint_stmt="This is related to the contents of the pyproject.toml file.",
                reference="invalid-pyproject-toml",
            )

        if metadata.license is None:
            raise ImproperProjectMetadata(
                message="No license is declared in `pyproject.toml`.",
                context=(
                    "It is required for this to be packaged using sphinx-theme-builder."
                ),
                hint_stmt="This is related to the contents of the pyproject.toml file.",
                reference="no-license-declared",
            )

        # Ensure that nothing other than the version is dynamic.
        metadata.version = version
        try:
            metadata.dynamic.remove("version")
        except ValueError:
            pass  # it doesn't exist, that's fine.

        if metadata.dynamic:
            raise ImproperProjectMetadata(
                message="Got unsupported keys for dynamic metadata.",
                context=str(metadata.dynamic),
                hint_stmt="This is related to the contents of the pyproject.toml file.",
                reference="unsupported-dynamic-keys",
            )

        # TODO: Factor this out, and add proper error messages.
        tool_config = pyproject_data.get("tool", {}).get("sphinx-theme-builder", {})

        # Get the node-version. This gets validated when executing a compilation.
        try:
            node_version = tool_config["node-version"]
        except KeyError:
            raise ImproperProjectMetadata(
                message="Did not get any node-version, from pyproject.toml file.",
                context=(
                    "It is required for this to be packaged using sphinx-theme-builder."
                ),
                hint_stmt=Text(
                    "Did you set node-version in [tool.sphinx-theme-builder]?"
                ),
                reference="no-node-version",
            )

        try:
            theme_name = tool_config["theme-name"]
        except KeyError:
            theme_name = kebab_name

        try:
            additional_compiled_static_assets = tool_config[
                "additional-compiled-static-assets"
            ]
        except KeyError:
            additional_compiled_static_assets = []

        return Project(
            kebab_name=kebab_name,
            metadata=metadata,
            location=path,
            theme_name=theme_name,
            node_version=node_version,
            additional_compiled_static_assets=additional_compiled_static_assets,
        )

    @property
    def snake_name(self) -> str:
        return self.kebab_name.replace("-", "_")

    @property
    def source_path(self) -> Path:
        return self.location / "src"

    @property
    def python_package_path(self) -> Path:
        return self.source_path / self.snake_name

    @property
    def theme_path(self) -> Path:
        return self.python_package_path / "theme" / self.theme_name

    @property
    def theme_static_path(self) -> Path:
        return self.theme_path / "static"

    @property
    def assets_path(self) -> Path:
        return self.python_package_path / "assets"

    @property
    def theme_conf_path(self) -> Path:
        return self.theme_path / "theme.conf"

    @property
    def input_scripts_path(self) -> Path:
        return self.assets_path / "scripts"

    @property
    def output_script_path(self) -> Path:
        return self.theme_static_path / "scripts" / (self.kebab_name + ".js")

    @property
    def input_stylesheets_path(self) -> Path:
        return self.assets_path / "styles"

    @property
    def output_stylesheet_path(self) -> Path:
        return self.theme_static_path / "styles" / (self.kebab_name + ".css")

    @property
    def compiled_assets(self) -> Tuple[str, ...]:
        """A sequence of compiled assets, as relative POSIX paths (to project root)."""
        compiled_assets = [
            self.output_script_path,
            self.output_stylesheet_path,
        ]
        compiled_assets.extend(
            [
                self.theme_static_path / asset_name
                for asset_name in self.additional_compiled_static_assets
            ]
        )

        files = tuple(
            path.relative_to(self.location).as_posix() for path in compiled_assets
        )
        # Tack on the .map files
        return files + tuple(map((lambda x: x + ".map"), files))

    def validate_file_structure_and_contents(self) -> None:
        """Validate the project's file structure and check that it matches the template.

        This is an important diagnostic step, to present useful and clear error messages
        to users who are transitioning from an existing theme to this one.
        """
        #
        # File presence
        #
        package_init_file = self.python_package_path / "__init__.py"
        assert package_init_file.exists(), (
            "This should've been validated in `Project.from_path`"
        )

        if not self.theme_conf_path.exists():
            raise InvalidProjectStructure(
                message=f"{self.theme_conf_path} does not exist.",
                context="It is required for using sphinx-theme-builder.",
                hint_stmt="Did you set the correct theme-name in pyproject.toml?",
                reference="missing-theme-conf",
            )

        if not self.input_scripts_path.exists():
            raise InvalidProjectStructure(
                message=f"{self.input_scripts_path} does not exist.",
                context="It is required for using sphinx-theme-builder.",
                hint_stmt=None,
                reference="missing-javascript",
            )

        if not self.input_stylesheets_path.exists():
            raise InvalidProjectStructure(
                message=f"{self.input_stylesheets_path} does not exist.",
                context="It is required for using sphinx-theme-builder.",
                hint_stmt=None,
                reference="missing-stylesheet",
            )

        #
        # File contents
        #
        expected_stylesheet = self.output_stylesheet_path.relative_to(
            self.theme_static_path
        ).as_posix()

        theme_conf_parser = configparser.ConfigParser()
        try:
            with self.theme_conf_path.open() as file:
                theme_conf_parser.read_file(file, source=file.name)
            specified_stylesheet = theme_conf_parser.get("theme", "stylesheet")
        except (OSError, configparser.Error) as error:
            raise InvalidProjectStructure(
                message=f"Could not open/parse {self.theme_conf_path}.",
                context=str(error),
                hint_stmt=None,
                reference="could-not-read-theme-conf",
            ) from error

        if specified_stylesheet != expected_stylesheet:
            raise InvalidProjectStructure(
                message=f"Incorrect stylesheet set in {self.theme_conf_path}.",
                context=f"Expected {expected_stylesheet}, got {specified_stylesheet}.",
                hint_stmt=None,
                reference="theme-conf-incorrect-stylesheet",
            )

    def get_metadata_file_contents(self) -> str:
        """Get contents for the `METADATA` file in a wheel."""
        return str(self.metadata.as_rfc822())

    def get_license_contents(self) -> str:
        """Get contents for the `LICENSE` file in a wheel."""
        return self.metadata.license.text

    def get_entry_points_contents(self) -> str:
        """Get contents for the `entry_points.txt` file in a wheel."""
        lines: List[str] = []
        for group, mapping in self.metadata.entrypoints.items():
            if lines:
                lines.append("")  # blank line, for visual clarity
            lines.append(f"[{group}]")
            for name, entrypoint in mapping.items():
                lines.append(f"{name} = {entrypoint}")
        return "\n".join(lines)



"""CLI passthrough, on pty-supporting systems.

Based on code licensed as:

    Copyright (c) 2015-2017, Terminal Labs
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
        * Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
        * Neither the name of Terminal Labs nor the
        names of its contributors may be used to endorse or promote products
        derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL TERMINAL LABS BE LIABLE FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Largely adapted from https://github.com/terminal-labs/cli-passthrough at commit
9337fbbaa59d12a1daaaba31a6dec9ce3107f8c7 (filename: cli_passthrough/_passthrough.py),
which in turn was adapted from StackOverflow.
"""

import sys
from typing import List

if sys.platform != "win32":
    import errno
    import fcntl
    import os
    import pty
    import shutil
    import struct
    import termios
    from itertools import chain
    from select import select
    from subprocess import Popen

    (_COLUMNS, _ROWS) = shutil.get_terminal_size(fallback=(80, 20))

    def _set_size(fd: int) -> None:
        """Found at: https://stackoverflow.com/a/6420070"""
        size = struct.pack("HHHH", _ROWS, _COLUMNS, 0, 0)
        fcntl.ioctl(fd, termios.TIOCSWINSZ, size)

    def passthrough_run(args: List[str]) -> int:
        """Largely found in https://stackoverflow.com/a/31953436"""
        masters, slaves = zip(pty.openpty(), pty.openpty())
        for fd in chain(masters, slaves):
            _set_size(fd)

        with Popen(args, stdin=sys.stdin, stdout=slaves[0], stderr=slaves[1]) as p:
            for fd in slaves:
                os.close(fd)  # no input
            readable = {
                masters[0]: sys.stdout.buffer,  # store buffers separately
                masters[1]: sys.stderr.buffer,
            }
            while readable:
                for fd in select(readable, [], [])[0]:
                    try:
                        data = os.read(fd, 1024)  # read available
                    except OSError as e:
                        if e.errno != errno.EIO:
                            # time to clean up!
                            p.kill()
                            for fd in masters:
                                os.close(fd)
                            raise
                        del readable[fd]  # EIO means EOF on some systems
                    else:
                        if not data:  # EOF
                            del readable[fd]
                        else:
                            readable[fd].write(data)
                            readable[fd].flush()
        for fd in masters:
            os.close(fd)
        return p.returncode


else:
    import subprocess

    def passthrough_run(args: List[str]) -> int:
        p = subprocess.run(args)
        return p.returncode
